#include <cmath>
#include <iostream>
#include <vector>
using namespace std;

inline int round_to_int(double x) { 
	return int(x + 0.5); 
}

/// Default values for \p eps for relative comparison in approx_eq.
const double default_eps_r = 1e-10; // QQQ: Change where comparing a and b because numerical problems occur even for large differences

/// Default values for \p eps for absolute comparison in approx_eq.
const double default_eps_a = 1e-10; // QQQ: Change where comparing a and b because numerical problems occur even for large differences

enum approx_tp { aprx_r, aprx_a };

/** Compares the floating point values and \p x and \p y and returns true only
    if they are closer to each other than

    - eps*min(x, y) (\p comp = aprx_r = relative comparison), or
    - eps           (\p comp = aprx_a = absolute comparison).

    If \p eps < 0 then \p default_eps_r respectively \p default_eps_a is used
    instead. */
inline bool 
approx_eq(double x, double y, approx_tp comp = aprx_r, 
          double eps = -1)
{
    using namespace std;
    if (comp == aprx_a) {
        if (eps < 0) 
            eps = default_eps_a;
        return abs(x - y) <= eps;
    } else {
        if (eps < 0) 
            eps = default_eps_r;
        return abs(x - y) <= eps * min(abs(x), abs(y));
    }
}

/** Set of disturbance and postestimation deviation and the parameters used to
    generate it. */
class s_vp_tp {
public:
    s_vp_tp(double, double, double, double, double, bool);
    inline double s();
    inline double vp();
    inline double a_0();
    inline double a_1();
    inline double b_0();
    inline double b_1();
    inline double beta_0();
    inline double beta_1();
    inline double gamma_0();
    inline double gamma_1();
private:
    void update_P(int m);
    void update_s();
    void update_vp();
    double s_mb; // Disturbance
    double vp_mb; // Postestimation deviation
    double a_mb[2];
    double b_mb[2];
    double beta_mb[2];
    double gamma_mb[2];
    double P_mb[2];
};

double s_vp_tp::s() { return s_mb; }
double s_vp_tp::vp() { return vp_mb; }
double s_vp_tp::a_0() { return a_mb[0]; }
double s_vp_tp::a_1() { return a_mb[1]; }
double s_vp_tp::b_0() { return b_mb[0]; }
double s_vp_tp::b_1() { return b_mb[1]; }
double s_vp_tp::beta_0() { return beta_mb[0]; }
double s_vp_tp::beta_1() { return beta_mb[1]; }
double s_vp_tp::gamma_0() { return gamma_mb[0]; }
double s_vp_tp::gamma_1() { return gamma_mb[1]; }

s_vp_tp::s_vp_tp(double a_0, double b_0, double beta_0, 
                 double beta_1, double gamma_0, bool switch_qqq)
{
    a_mb[0] = a_0;
    b_mb[0] = b_0;
    beta_mb[0] = beta_0;
    beta_mb[1] = beta_1;
    gamma_mb[0] = gamma_0;

    // QQQ: this is only one possibility (there are others that shouldbe
    // investigated as well):
    if (!switch_qqq) {
        a_mb[1] = sqrt(1-pow(a_mb[0],2));
        b_mb[1] = sqrt(1-pow(b_mb[0],2));
        gamma_mb[1] = gamma_mb[0];
    } else {
        a_mb[1] = sqrt(1-pow(b_mb[0],2));
        b_mb[1] = sqrt(1-pow(a_mb[0],2));
        gamma_mb[1] = 1 - gamma_mb[0];
    }

    for (int m = 0; m <= 1; ++m)
        update_P(m);
    update_s();
    update_vp();
}

/** Updates \p P_mb[m]. For calculation the current values of \p a_mb[m],
    \p b_mb[m], etc. are used. */
void s_vp_tp::update_P(int m) {
    double a = a_mb[m], b = b_mb[m], beta = beta_mb[m];
    double &P = P_mb[m];
    if (approx_eq(a, 0, aprx_a))
        P = 0.5 * pow(b,2) * pow(beta-1,2);
    else if (approx_eq(b, 0, aprx_a))
        P = 0.5 * pow(a,2) * pow(beta,2);
    else if (approx_eq(a, b, aprx_a))
        P = 1./3. * pow(a,2);
    else
        P = 
            1. / (2.*pow(pow(a,2)-pow(b,2),2)) * 
            (pow(a,6) * pow(beta,2) + 
             pow(b,6) * pow(1-beta,2) +
             pow(a,4) * pow(b,2) * beta * (4-7*beta) +
             pow(a,2) * pow(b,4) * (1-beta) * (7*beta-3)) +
            2.*pow(a,4)*pow(b,4) / pow(pow(a,2)-pow(b,2),3) *
            (1 - 6*beta*(1-beta)) * log(a/b);
}

/** Updates the disturbance \p s_mb. */
void s_vp_tp::update_s() {
    s_mb = 0.;
    for (int m = 0; m <= 1; ++m) {
        double a = a_mb[m], b = b_mb[m], beta = beta_mb[m], gamma = gamma_mb[m];
        s_mb += P_mb[m];
        s_mb -= 1./3. * 
            (pow(a,2)*beta*(1+gamma) +
             pow(b,2)*(1-beta)*(2-gamma) +
             2*a*b*sqrt(beta*(1-beta))*sqrt(gamma*(1-gamma)));
    }
    s_mb += 1./3.;
}

/** Updates the postestimation deviation \p vp_mb. */
void s_vp_tp::update_vp() {
    vp_mb = 0.;
    for (int m = 0; m <= 1; ++m) {
        double a = a_mb[m], b = b_mb[m], beta = beta_mb[m];
        vp_mb += P_mb[m];
        if (!(approx_eq(a, 0, aprx_a) && approx_eq(b, 0, aprx_a)))
            vp_mb -= 0.5 *
                pow(pow(a,2)*beta + pow(b,2)*(1-beta), 2) / (pow(a,2)+pow(b,2));
    }
}

int main() {
    // Check why this doesn't give 1/6 (probably because of the optimized kappa):
    s_vp_tp s_vp1a(1/*a_0*/, 1/*b_0*/, 0/*beta_0*/, 0/*beta_1*/, 0/*gamma_0*/, true);
    cerr << "s=" << s_vp1a.s() << ", vp=" << s_vp1a.vp() << endl;
    s_vp_tp s_vp1b(1/*a_0*/, 1/*b_0*/, 0/*beta_0*/, 0/*beta_1*/, 0/*gamma_0*/, false);
    cerr << "s=" << s_vp1b.s() << ", vp=" << s_vp1b.vp() << endl;

    // Check why this doesn't give 1/6 (probably because of the optimized kappa):
    s_vp_tp s_vp2a(0/*a_0*/, 1/*b_0*/, 0.5/*beta_0*/, 0.5/*beta_1*/, 0.5/*gamma_0*/, true);
    cerr << "s=" << s_vp2a.s() << ", vp=" << s_vp2a.vp() << endl;
    s_vp_tp s_vp2b(0/*a_0*/, 1/*b_0*/, 0.5/*beta_0*/, 0.5/*beta_1*/, 0.5/*gamma_0*/, false);
    cerr << "s=" << s_vp2b.s() << ", vp=" << s_vp2b.vp() << endl;

    s_vp_tp s_vp3a(0.70710678118655/*a_0*/, 0.70710678118655/*b_0*/, 0.5/*bea_0*/, 0.5/*bea_1*/, 0.5/*gamma_0*/, true);
    cerr << "s=" << s_vp3a.s() << ", vp=" << s_vp3a.vp() << endl;
    s_vp_tp s_vp3b(0.70710678118655/*a_0*/, 0.70710678118655/*b_0*/, 0.5/*beb_0*/, 0.5/*beb_1*/, 0.5/*gamma_0*/, true);
    cerr << "s=" << s_vp3b.s() << ", vp=" << s_vp3b.vp() << endl;

    s_vp_tp s_vp4a(0/*a_0*/, 1/*b_0*/, 0.5/*bea_0*/, 0.5/*bea_1*/, 0.5/*gamma_0*/, true);
    cerr << "s=" << s_vp4a.s() << ", vp=" << s_vp4a.vp() << endl;
    s_vp_tp s_vp4b(0/*a_0*/, 1/*b_0*/, 0.5/*beb_0*/, 0.5/*beb_1*/, 0.5/*gamma_0*/, false);
    cerr << "s=" << s_vp4b.s() << ", vp=" << s_vp4b.vp() << endl;

    s_vp_tp s_vp5a(0/*a_0*/, 0/*b_0*/, 0/*bea_0*/, 0/*bea_1*/, 0/*gamma_0*/, true);
    cerr << "s=" << s_vp5a.s() << ", vp=" << s_vp5a.vp() << endl;
    s_vp_tp s_vp5b(0/*a_0*/, 0/*b_0*/, 0/*beb_0*/, 0/*beb_1*/, 0/*gamma_0*/, false);
    cerr << "s=" << s_vp5b.s() << ", vp=" << s_vp5b.vp() << endl;

    s_vp_tp s_vp6a(0.4/*a_0*/, 0.2/*b_0*/, 0.6/*bea_0*/, 0.1/*bea_1*/, 0.8/*gamma_0*/, true);
    cerr << "s=" << s_vp6a.s() << ", vp=" << s_vp6a.vp() << endl;
    s_vp_tp s_vp6b(0.4/*a_0*/, 0.2/*b_0*/, 0.6/*bea_0*/, 0.1/*bea_1*/, 0.8/*gamma_0*/, false);
    cerr << "s=" << s_vp6b.s() << ", vp=" << s_vp6b.vp() << endl;

    size_t nsteps = 100, maxstep = nsteps-1;
    double tmp = 0.;
    vector<s_vp_tp*> min_s_vps(1000, (s_vp_tp*)0);
    for (size_t i = 0; i <= maxstep; ++i) {
        cerr << "QQQ1" << endl;
        for (size_t j = 0; j <= maxstep; ++j)
            for (size_t k = 0; k <= maxstep; ++k)
                for (size_t l = 0; l <= maxstep; ++l)
                    for (size_t m = 0; m <= maxstep; ++m)
                        for (int n = 0; n <= 0/*QQQ:1*/; ++n) { // QQQ: maybe this is superfluous because of symmetry?
                            double a_0 = 1.*i/maxstep;
                            double b_0 = 1.*j/maxstep;
                            double beta_0 = 1.*k/maxstep;
                            double beta_1 = 1.*l/maxstep;
                            double gamma_0 = 1.*m/maxstep;
                            s_vp_tp *s_vp = new s_vp_tp(a_0, b_0, beta_0, beta_1, gamma_0, n);
                            int slot = round_to_int(s_vp->vp()*(min_s_vps.size()-1));
                            if (min_s_vps[slot] == 0) {
                                min_s_vps[slot] = s_vp;
                            } else if (s_vp->s() < min_s_vps[slot]->s()) {
                                delete min_s_vps[slot];
                                min_s_vps[slot] = s_vp;
                            } else
                                delete s_vp;
                        }
    }

    cout << "# slot vp s a_0 a_1 b_0 b_1 beta_0 beta_1 gamma_0 gamma_1" << endl;
    for (size_t i = 0; i <= min_s_vps.size(); ++i)
        if (min_s_vps[i] != 0)
            cout << i << "\t"
                 << min_s_vps[i]->vp() << "\t"
                 << min_s_vps[i]->s() << "\t"
                 << min_s_vps[i]->a_0() << "\t"
                 << min_s_vps[i]->a_1() << "\t"
                 << min_s_vps[i]->b_0() << "\t"
                 << min_s_vps[i]->b_1() << "\t"
                 << min_s_vps[i]->beta_0() << "\t"
                 << min_s_vps[i]->beta_1() << "\t"
                 << min_s_vps[i]->gamma_0() << "\t"
                 << min_s_vps[i]->gamma_1()
                 << endl;

    for (size_t i = 0; i <= maxstep; ++i)
        delete min_s_vps[i];

    return 0;
}
